use tokio::sync::Mutex;

use super::domain::repository::AdminRepositoryWrapper;
use super::domain::{AdminApp, AdminModel, AdminModelConfig, AdminRepository, DynAdminRepository};
use crate::service::templates::Templates;
use std::collections::HashMap;
use std::sync::Arc;

pub trait AdminState {
    fn get_templates(&self) -> &Templates;
    fn get_registry(&self) -> SharedAdminRegistry;
}

pub type SharedAdminRegistry = Arc<AdminRegistry>;

// main registry.
pub struct AdminRegistry {
    base_path: String,
    apps: HashMap<String, internal::AdminApp>,
    models: HashMap<String, internal::AdminModel>,
    repositories: HashMap<String, Arc<Mutex<dyn DynAdminRepository>>>,
}

impl AdminRegistry {
    pub fn new(base_path: &str) -> Self {
        AdminRegistry {
            base_path: base_path.to_owned(),
            apps: HashMap::new(),
            models: HashMap::new(),
            repositories: HashMap::new(),
        }
    }

    pub fn get_apps(&self) -> Vec<AdminApp> {
        self.apps
            .iter()
            .map(|(key, node)| self.get_app(key, node))
            .collect()
    }

    fn get_app(&self, key: &str, node: &internal::AdminApp) -> AdminApp {
        let my_models = self.get_models(key);
        AdminApp {
            name: key.to_owned(),
            key: node.name.to_owned(),
            app_url: format!("/{}/app/{}", self.base_path, key.to_owned()),
            models: my_models,
        }
    }

    pub fn register_app(&mut self, name: &str) -> String {
        let key = self.get_key(name);
        self.apps.insert(
            key.to_owned(),
            internal::AdminApp {
                key: key.to_owned(),
                name: name.to_owned(),
            },
        );
        key
    }

    fn get_key(&self, name: &str) -> String {
        slug::slugify(name)
    }

    fn model_from_internal(&self, internal_model: &internal::AdminModel) -> AdminModel {
        let admin_url = format!(
            "/{}/app/{}/model/{}",
            self.base_path, internal_model.app_key, internal_model.model_key
        );
        AdminModel {
            key: internal_model.model_key.clone(),
            name: internal_model.name.clone(),
            view_only: false,
            add_url: Some(format!("{}/add", admin_url)),
            admin_url: admin_url,
        }
    }

    pub fn get_models(&self, app_key: &str) -> Vec<AdminModel> {
        self.models
            .iter()
            .filter(|(key, _)| key.starts_with(&format!("{}.", app_key)))
            .map(|(_, model)| self.model_from_internal(model))
            .collect()
    }

    pub fn get_model(&self, app_key: &str, model_key: &str) -> Option<AdminModel> {
        let full_model_key = format!("{}.{}", app_key, model_key);
        let internal_model = self.models.get(&full_model_key)?;

        // unfinished: we need to think about model_key vs. model_id vs. entry_id, as "name" is ambiguous.
        Some(self.model_from_internal(internal_model))
    }

    fn register_model_config(&mut self, model: AdminModelConfig) -> Result<String, String> {
        let local_config = internal::AdminModel::from(model);
        if local_config.model_key.is_empty() {
            return Err("No model name".to_owned());
        }
        let local_config_name = format!("{}.{}", local_config.app_key, local_config.model_key);
        if self.models.contains_key(&local_config_name) {
            return Err(format!("Model {} already exists", local_config_name));
        }
        let full_model_key = local_config_name.clone();
        self.models.insert(local_config_name, local_config);
        Ok(full_model_key)
    }

    pub fn register_model<R: AdminRepository + 'static>(
        &mut self,
        model: AdminModelConfig,
        repository: R,
    ) -> Result<(), String> {
        let model_key = self.register_model_config(model)?;
        let repository = AdminRepositoryWrapper::new(repository);
        self.repositories
            .insert(model_key, Arc::new(Mutex::new(repository)));
        Ok(())
    }

    pub fn get_repository(
        &self,
        app_key: &str,
        model_key: &str,
    ) -> Result<Arc<Mutex<dyn DynAdminRepository>>, String> {
        let full_model_key = format!("{}.{}", app_key, model_key);
        if let Some(repo) = self.repositories.get(&full_model_key) {
            // Clone the Arc to return a reference to the repository
            return Ok(Arc::clone(repo));
        } else {
            return Err("Couldn't find repository".to_owned());
        }
    }
}

mod internal {
    // how the registry saves data internally.

    use super::super::domain::AdminModelConfig;
    #[derive(Clone)]
    pub struct AdminApp {
        pub key: String,
        pub name: String,
    }

    #[derive(Clone)]
    pub struct AdminModel {
        pub app_key: String,
        pub model_key: String,
        pub name: String,
    }

    impl From<AdminModelConfig> for AdminModel {
        fn from(value: AdminModelConfig) -> Self {
            AdminModel {
                app_key: value.app_key,
                model_key: slug::slugify(value.name.clone()),
                name: value.name,
            }
        }
    }

    impl From<(&str, &str)> for AdminModel {
        fn from(value: (&str, &str)) -> Self {
            AdminModel {
                app_key: value.0.to_owned(),
                model_key: slug::slugify(value.1.to_owned()),
                name: value.1.to_owned(),
            }
        }
    }
}
